package baselinecomparison;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.text.SimpleDateFormat;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTable;
import javax.swing.SwingUtilities;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableColumn;

import com.wily.introscope.console.thornhill.manager.IThornhillItem;
import com.wily.introscope.console.thornhill.manager.agent.AgentManager;
import com.wily.introscope.console.thornhill.manager.agent.ThornhillAgent;
import com.wily.introscope.console.thornhill.manager.agent.ThornhillMetric;
import com.wily.introscope.console.thornhill.manager.agent.ThornhillMetricPath;
import com.wily.introscope.console.thornhill.manager.clock.ClockManager;
import com.wily.introscope.console.thornhill.ui.ThornhillUserInterface;
import com.wily.introscope.spec.metric.AgentName;
import com.wily.introscope.spec.metric.IncompatibleTypesException;
import com.wily.introscope.spec.metric.InvalidRegexException;
import com.wily.introscope.spec.metric.Metric;
import com.wily.introscope.spec.metric.MetricData;
import com.wily.introscope.spec.metric.RegularExpressionBuilder;
import com.wily.introscope.spec.server.AgentMetricSpecifier;
import com.wily.introscope.spec.server.AgentNameSpecifierIgnoringDomain;
import com.wily.introscope.spec.server.DomainSpecifier;
import com.wily.introscope.spec.server.HistoricalAgentName;
import com.wily.introscope.spec.server.api.IAgentMetricSpecifier;
import com.wily.introscope.spec.server.api.IAgentNameSpecifierIgnoringDomain;
import com.wily.introscope.spec.server.api.IDomainSpecifier;
import com.wily.introscope.spec.server.api.IMetricSpecifier;
import com.wily.introscope.spec.server.beans.metricdata.IMetricDataManager;
import com.wily.introscope.spec.server.beans.metricdata.IMetricDataValue;
import com.wily.introscope.spec.server.beans.metricdata.ITimeslicedResultSet;
import com.wily.introscope.spec.server.beans.metricdata.MetricDataQuery;
import com.wily.introscope.spec.server.beans.metricgroup.specifiers.AnyNumericalAgentMetricSpecifier;
import com.wily.introscope.spec.server.beans.metricgroup.specifiers.BooleanAndAgentMetricSpecifier;
import com.wily.introscope.spec.server.beans.metricgroup.specifiers.MetricSpecifier;
import com.wily.introscope.stat.gatherer.GathererFactory;
import com.wily.introscope.stat.gatherer.IGatherer;
import com.wily.introscope.stat.timeslice.ATimeslicedValue;
import com.wily.introscope.util.HistoricalQueryRange;
import com.wily.isengard.messageprimitives.ConnectionException;
import com.wily.ui.swing.filechooser.WilyFileChooser;
import com.wily.ui.swing.table.TableSorter_v2;
import com.wily.ui.swing.table.TooltipSavyJTable;
import com.wily.util.datetime.DateFormatException;
import com.wily.util.feedback.IModuleFeedbackChannel;
import com.wily.util.feedback.Module;

public class BaselineComparisonTypeViewer extends
		com.wily.introscope.console.thornhill.ui.typeviewer.ATypeViewer
	{
	
	static 
		{
		sDateFormat = new SimpleDateFormat("MMMMM dd, yyyy");	
		sDateAndTimeFormat = new SimpleDateFormat("MMMMM dd, yyyy hh:mm a z");
		}
	
		public JPanel fPanel = new JPanel();
		public static Module sModule = new Module("Baseline"); 
		public static IModuleFeedbackChannel sChannel = null;
		public TooltipSavyJTable fTable;
		public DefaultTableModel fModel;

		public static BaselineComparisonPicker sEarlierPicker;
		public static BaselineComparisonPicker sLaterPicker;
		public HashMap<String, IGatherer> fEarlyGatherers = new HashMap<String, IGatherer>();
		public HashMap<String, IGatherer> fLateGatherers = new HashMap<String, IGatherer>();
		public static SimpleDateFormat sDateFormat;
		public static SimpleDateFormat sDateAndTimeFormat;
		public static String kDefaultFile = "metrics.csv";

		public void displayItem(final IThornhillItem item, HistoricalQueryRange range)
				throws ConnectionException
			{
			if(sChannel == null)
				{
				sChannel = getTypeViewerHelper().getFeedback();
				}
			
			sChannel.debug(sModule, "ENTER displayItem - " + item.getFullName());
			
			final AgentName[] agents = initAgentNames();
			final ClockManager clockManager = getUserInterfaceServices().getAllManagersHandle().getClockManager();
			
			SwingUtilities.invokeLater(new Runnable() 
				{

				@SuppressWarnings("deprecation")
				public void run()
					{
					fPanel.removeAll();
					
					JPanel topPanel = new JPanel();
					
					FlowLayout flowLayout = new FlowLayout();
					flowLayout.setHgap(25);
					flowLayout.setVgap(5);
					topPanel.setLayout(flowLayout);
					long now = System.currentTimeMillis();
					if(sEarlierPicker == null)
						{
						ThornhillUserInterface thornhillUserInterface = getUserInterfaceServices().getThornhillUserInterface();
							
						sEarlierPicker = new BaselineComparisonPicker(item, getTypeViewerHelper().getGenericServices(), clockManager, thornhillUserInterface,"EARLIER TIME RANGE", agents, new Date(now - (1000 * 60 * 60)));
					
						if(sLaterPicker == null)
							{
							sLaterPicker = new BaselineComparisonPicker(item, getTypeViewerHelper().getGenericServices(), clockManager, thornhillUserInterface, "LATER TIME RANGE", agents, new Date((now - (1000 * 60 * 60)) + (8 * 60 * 1000)));
							sLaterPicker.getDurationPicker().setEnabled(false);
							}
						
						sEarlierPicker.getDurationPicker().addActionListener(new ActionListener() {

						public void actionPerformed(ActionEvent event)
							{
							sLaterPicker.getDurationPicker().setSelectedIndex(sEarlierPicker.getDurationPicker().getSelectedIndex());
							}
					
					});						}
					
					topPanel.add(sEarlierPicker);
					
					JPanel buttonPanel = new JPanel();
					BorderLayout borderLayout = new BorderLayout();
					borderLayout.setVgap(5);
					buttonPanel.setLayout(borderLayout);
					
					JButton goButton = new JButton(">> GO <<");
					goButton.addActionListener(new ActionListener() {
						
						public void actionPerformed(ActionEvent arg0)
							{
							while(fModel.getRowCount() > 0)
								{
								fModel.removeRow(0);
								}
							
							fTable.removeAll();
							
							if(fEarlyGatherers != null)
								{
								fEarlyGatherers.clear();
								}
							
							if(fLateGatherers != null)
								{
								fLateGatherers.clear();
								}
							
							queryAndDisplay(item);
							}
					});
					
					JButton exportButton = new JButton("<< EXPORT >>");
					
					exportButton.addActionListener(new ActionListener() {


						public void actionPerformed(ActionEvent arg0)
							{
							WilyFileChooser chooser = new WilyFileChooser();
							chooser.setMultiSelectionEnabled(false);
							
							chooser.setSelectedFile(new File (kDefaultFile));
							
							
							int returnVal = chooser.showDialog(BaselineComparisonTypeViewer.this.getComponent(), "Save");

							File file;
							if(returnVal == JFileChooser.CANCEL_OPTION)
								{
								return;
								}
							else if (returnVal == JFileChooser.APPROVE_OPTION)
								{
								file = chooser.getSelectedFile();
								}
							else
								{
								file = new File(kDefaultFile);
								}

							sChannel.debug(sModule, file.getName());
							
							FileOutputStream fos = null;
							try
								{
								fos = new FileOutputStream(file);
								} 
							catch (FileNotFoundException e)
								{
								sChannel.error(e);
								}
							
							if(fos == null)
								{
								return;
								}
								
							BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(fos));
							
							String newLine = System.getProperty("line.separator");
							String line = "";
							for(int i = 0; i < BaselineComparisonTable.sColumns.length; i++)
								{
								line += BaselineComparisonTable.sColumns[i];
								
								if(i+1 < BaselineComparisonTable.sColumnColors.length)
									{
									line += ",";
									}
								}
							try
								{
								writer.write(line + newLine); // print columns
								if(fTable != null)
									{
									for(int i = 0; i < fModel.getRowCount(); i++)
										{
										line = "";
										for (int j = 0; j < BaselineComparisonTable.sColumns.length; j++)
											{
											line += fModel.getValueAt(i, j);
											
											if(j+1 < BaselineComparisonTable.sColumns.length)
												{
												line += ",";
												}
											}
										writer.write(line + newLine);
										}
						
									writer.write(newLine);
									}
								
								writer.flush();
								writer.close();
								} 
							catch (IOException e)
								{
								sChannel.error(e);
								}
							}
					
					});
					
					goButton.setPreferredSize(new Dimension(100, 22));
					exportButton.setPreferredSize(new Dimension(100, 22));
					
					buttonPanel.add(goButton, BorderLayout.NORTH);
					buttonPanel.add(exportButton, BorderLayout.SOUTH);
					
					topPanel.add(buttonPanel);

					topPanel.add(sLaterPicker);
					
					JScrollPane topScrollPanel = new JScrollPane(topPanel, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
					
					JPanel bottomPanel = new JPanel();
					bottomPanel.setLayout(new BorderLayout());
					//bottomPanel.setLayout(new FlowLayout());
					
					fModel = new DefaultTableModel();
				
					fModel.setColumnIdentifiers(BaselineComparisonTable.sColumns);
					
					TableSorter_v2 sorter = new TableSorter_v2(fModel);
					sorter.sortByColumn(0);
					fTable = new BaselineComparisonTable(sorter);
					initTable(fTable);
					fTable.setPreferredSize(new Dimension(1024, 768));
										
					JScrollPane scrollTable = new JScrollPane(fTable, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
					scrollTable.setBorder(BorderFactory.createLoweredBevelBorder());
					scrollTable.setSize(new Dimension(1024, 768));
					bottomPanel.add(scrollTable, BorderLayout.CENTER);
					
					final JSplitPane splitPane = new JSplitPane(JSplitPane.VERTICAL_SPLIT, topScrollPanel, bottomPanel);
					splitPane.addComponentListener(new ComponentAdapter() {
					
						public void
						componentShown(ComponentEvent ce)
							{
							splitPane.setDividerLocation(0.3);
							}
						});
					
					fPanel.setLayout(new BorderLayout());
					fPanel.add(splitPane, BorderLayout.CENTER);
					fPanel.setPreferredSize(new Dimension(1600, 1200));
					}
			});
			
			sChannel.debug(sModule, "EXIT displayItem - " + item.getFullName());	
			}

		public AgentName[] initAgentNames()
			{
			List<ComparableProxyAgentName> agentNames = new LinkedList<ComparableProxyAgentName>();
			
			// Agents
			AgentManager agentManager = getUserInterfaceServices().getAllManagersHandle().getAgentManager();
			try
				{
				HistoricalAgentName[] availableHistoricalProcessesOnServer = agentManager.getAvailableHistoricalProcessesOnServer();
				for(int i = 0; i < availableHistoricalProcessesOnServer.length; i++)
					{
					ComparableProxyAgentName agentName = new ComparableProxyAgentName (availableHistoricalProcessesOnServer[i].getAgentName());
					agentNames.add(agentName);
					}
				} 
			catch (Exception e)
				{
				sChannel.error(e);
				} 

			Collections.sort(agentNames);
						
			AgentName[] result = new AgentName[agentNames.size()];
			for(int i = 0; i < result.length; i++)
				{
				ComparableProxyAgentName comparableProxyAgentName = agentNames.get(i);
				result[i] = comparableProxyAgentName.getAgentName();
				}
			
			return result;
			}

		public Component getComponent()
			{
			System.out.println("getComponent()");
			return fPanel;
			}

		public String getName()
			{
			return "Comparison";
			}

		public boolean isSupported(IThornhillItem item)
			{
			boolean result = false;
			if(item instanceof ThornhillAgent || item instanceof ThornhillMetricPath || item instanceof ThornhillMetric)
				{
				result = true;
				}
			return result;
			}

		public void initTable(JTable table)
			{
			table.setAutoCreateColumnsFromModel(false);

			TableColumn column = table.getColumnModel().getColumn(0);
			column.setPreferredWidth(600);
			column = table.getColumnModel().getColumn(1);
			column.setPreferredWidth(175);
			column = table.getColumnModel().getColumn(2);
			column.setPreferredWidth(175);
			column = table.getColumnModel().getColumn(3);
			column.setPreferredWidth(175);
			column = table.getColumnModel().getColumn(4);
			column.setPreferredWidth(175);
			column = table.getColumnModel().getColumn(5);
			column.setPreferredWidth(175);
			column = table.getColumnModel().getColumn(6);
			column.setPreferredWidth(175);
			column = table.getColumnModel().getColumn(7);
			column.setPreferredWidth(175);
			column = table.getColumnModel().getColumn(8);
			column.setPreferredWidth(175);
			column = table.getColumnModel().getColumn(9);
			column.setPreferredWidth(175);
			column = table.getColumnModel().getColumn(10);
			column.setPreferredWidth(175);
			column = table.getColumnModel().getColumn(11);
			column.setPreferredWidth(175);
			column = table.getColumnModel().getColumn(12);
			column.setPreferredWidth(175);
			}
		
	public void
	queryAndDisplay(IThornhillItem item)
		{
		try {
		
			AgentName earlierAgentName = (AgentName) sEarlierPicker.getAgentPicker().getSelectedItem();
			long durationInMillis = parseDuration((String) sEarlierPicker.getDurationPicker().getSelectedItem());
			Date earlierDate = sEarlierPicker.getDateTimeSelector().getDate();
			
			AgentName laterAgentName = (AgentName) sLaterPicker.getAgentPicker().getSelectedItem();
			Date laterDate = sLaterPicker.getDateTimeSelector().getDate();

			if ( earlierDate == null || laterDate == null)
				{
				String warningMessage = "Error specifying date and time: ";
				if(earlierDate == null)
					{
					warningMessage += " earlier date is unreadable; ";
					}
				if(laterDate == null)
					{
					warningMessage += " later date is unreadable";
					}
				
				warningMessage += sNewLine + " Is there a space between your time digits and the AM/PM mark?";
				warningMessage += sNewLine + "Correct is > 11:15 AM  <  and not  > 11:15AM < ";
				
				JOptionPane.showMessageDialog(BaselineComparisonTypeViewer.this.getComponent(), warningMessage );
				return;
				}
			
			IMetricDataManager metricDataManager = getTypeViewerHelper().getMetricDataManager();

			HistoricalQueryRange earlierTimeRange = new HistoricalQueryRange(HistoricalQueryRange.kTimeAbsolute, earlierDate.getTime(), HistoricalQueryRange.kTimeAbsolute, earlierDate.getTime() + durationInMillis); 
			IMetricSpecifier specifier = getQuerySpecifier(item, earlierAgentName);			
			MetricDataQuery query = new MetricDataQuery(specifier, earlierTimeRange);
			
			Date eDate = new Date(earlierTimeRange.getStartTime());
			
			sChannel.debug(sModule, "earlierRange from " + sDateAndTimeFormat.format(eDate) + " for " + earlierTimeRange.getDurationMinutes() + " minutes");
			
			ITimeslicedResultSet[] earlierQueryResultSet = metricDataManager.executeSyncQuery(getTypeViewerHelper().getSessionToken(), query);
			aggregateResultSet(earlierQueryResultSet, true);
			
			HistoricalQueryRange laterTimeRange = new HistoricalQueryRange(HistoricalQueryRange.kTimeAbsolute, laterDate.getTime(), HistoricalQueryRange.kTimeAbsolute, laterDate.getTime() + durationInMillis); 
			specifier = getQuerySpecifier(item, laterAgentName);
			MetricDataQuery laterQuery = new MetricDataQuery(specifier, laterTimeRange);
			
			Date lDate = new Date(laterTimeRange.getStartTime());
			sChannel.debug(sModule, "laterRange from " + sDateAndTimeFormat.format(lDate) + " for " + laterTimeRange.getDurationMinutes() + " minutes");
			
			ITimeslicedResultSet[] laterQueryResultSet = metricDataManager.executeSyncQuery(getTypeViewerHelper().getSessionToken(), laterQuery);
			aggregateResultSet(laterQueryResultSet, false);

			informOfResultSetSizes(item, earlierQueryResultSet, laterQueryResultSet);
			// replace with Data/No Data visual feedback.
			
			if(earlierQueryResultSet.length <= 0)
				{
				sChannel.debug(sModule, "EMPTY early result set");
				}
			else if(laterQueryResultSet.length <= 0)
				{
				sChannel.debug(sModule, "EMPTY later result set");
				}

			//printResultSet(laterQueryResultSet);			
		
			

			addResultToTable(earlierDate, laterDate, durationInMillis);
			}
		catch(Exception e)
			{
			sChannel.error(e);
			}
		
		}

	public static String sNewLine = System.getProperty("line.separator");
	
	public void informOfResultSetSizes(IThornhillItem item, ITimeslicedResultSet[] earlierQueryResultSet, ITimeslicedResultSet[] laterQueryResultSet) throws DateFormatException, InvalidRegexException
		{		
		String newLine = System.getProperty("line.separator");
		
		AgentName agentName = (AgentName) sEarlierPicker.getAgentPicker().getSelectedItem();
		Date date = sEarlierPicker.getDateTimeSelector().getDate();
		String duration = (String) sEarlierPicker.getDurationPicker().getSelectedItem();		
		String earlyMessage = "|| Queried " + getQuerySpecifier(item, agentName) + " from time " + sDateAndTimeFormat.format(date) + " for " + duration + " : " + earlierQueryResultSet.length; 
		
		agentName = (AgentName) sLaterPicker.getAgentPicker().getSelectedItem();
		date = sLaterPicker.getDateTimeSelector().getDate();
		duration = (String) sLaterPicker.getDurationPicker().getSelectedItem();
		String laterMessage = "|| Queried " + getQuerySpecifier(item, agentName) + " from time " + sDateAndTimeFormat.format(date) + " for " + duration + " : " + laterQueryResultSet.length;
		
		
		String message = "Query Results Summary";
		
		int size = 10;
		if(earlyMessage.length() > laterMessage.length())
			{
			size = earlyMessage.length();
			}
		else 
			{
			size = laterMessage.length();
			}
		
		int headerSize = ((size - message.length()) / 5) ;
		
		String headerPrefixSuffix = "";
		for (int i = 0; i < headerSize; i ++)
			{
			headerPrefixSuffix += "=";
			}
		message = headerPrefixSuffix + message + headerPrefixSuffix + newLine;
		
		int n = earlyMessage.length() - size;
		if (n > 0)
			{
			for(int i = 0; i < n; i++)
				{
				earlyMessage = earlyMessage + " ";
				
				for(int j = 0 ; j < earlierQueryResultSet.length; j++)
					{
					Metric[] metricNames = earlierQueryResultSet[j].getMetricNames();
					for (int k = 0; k < metricNames.length; k++)
						{
						System.out.println("  - " + metricNames[k].getMetricUrlWithoutDomain() + " ");
						}
					System.out.println("     . Period: " +  earlierQueryResultSet[j].getPeriod());
					System.out.println("     . Count: " + earlierQueryResultSet[j].getCount());
					System.out.println("     . Time: " + sDateAndTimeFormat.format(new Date(earlierQueryResultSet[j].getTimeMarker().getExactTimestampInMillis())));
					}
				}		
			}
		else
			{
			n = laterMessage.length() - size;
			for(int i = 0; i < 0; i ++)
				{
				laterMessage += " ";
				}
			}
		message += earlyMessage + newLine;
		message += laterMessage + newLine;
		message += newLine;
		for(int i = 1; i < size; i++)
			{
			message += "=";
			}
		message += newLine + newLine;
		message += "Too many zeros for result sets?  Try rolling further back in time.";
		
		JOptionPane.showMessageDialog(this.getComponent(), message, "Data Query Summary", JOptionPane.INFORMATION_MESSAGE);
		}

	public void addResultToTable(Date earlierDate, Date laterDate, long durationInMillis)
		{
		Set<String> keySet = fEarlyGatherers.keySet();
		Iterator<String> itr = keySet.iterator();

		while(itr.hasNext())
			{
			String next = itr.next();

			IGatherer earlyGatherer = fEarlyGatherers.get(next);
			IGatherer lateGatherer = fLateGatherers.get(next);
			
			if(earlyGatherer == null)
				{
				sChannel.debug("skipping row for lack of early data on metric: " + next);
				continue;
				}
			else if(lateGatherer == null)
				{
				sChannel.debug("skipping row for lack of late data on metric: " + next);
				continue;
				}
			
			ATimeslicedValue[] startValues = earlyGatherer.IGatherer_harvestAndReset(earlierDate.getTime() + durationInMillis);
			ATimeslicedValue[] endValues = lateGatherer.IGatherer_harvestAndReset(laterDate.getTime() + durationInMillis);
			
			sChannel.debug(sModule, "startValues.length = " + startValues.length);
			sChannel.debug(sModule, "endValues.length = " + endValues.length);

			ATimeslicedValue startValue = startValues[0];
			ATimeslicedValue endValue = endValues[0];
			
			String[] columns = new String[fTable.getColumnCount()];
			
			double defaultDouble = -1;
			columns[0] = next;
			columns[1] = "" + startValue.getValueAsDouble(defaultDouble);
			columns[2] = "" + endValue.getValueAsDouble(defaultDouble);
			columns[3] = "" + (((startValue.getValueAsDouble(defaultDouble) * 100) / endValue.getValueAsDouble(defaultDouble)) - 100);

			columns[4] = "" + startValue.getMinAsDouble(defaultDouble);
			columns[5] = "" + endValue.getMinAsDouble(defaultDouble);
			columns[6] = "" + (((startValue.getMinAsDouble(defaultDouble) * 100) / endValue.getMinAsDouble(defaultDouble)) - 100);
			
			columns[7] = "" + startValue.getMaxAsDouble(defaultDouble);
			columns[8] = "" + endValue.getMaxAsDouble(defaultDouble);
			columns[9] = "" + (((startValue.getMaxAsDouble(defaultDouble) * 100) / endValue.getMaxAsDouble(defaultDouble)) - 100);
			
			columns[10] = "" + startValue.getDataPointCount();
			columns[11] = "" + endValue.getDataPointCount();
			columns[12] = "" + (((startValue.getDataPointCount() * 100) / endValue.getDataPointCount()) - 100);
			
			sChannel.debug(sModule, "adding row for metric " + columns[0]);
			sChannel.debug(sModule, "columns in model: " + fModel.getColumnCount() + " columns in row: " + columns.length);
			fModel.addRow(columns);
			}
		
		fModel.fireTableDataChanged();
		}

	public void aggregateResultSet(ITimeslicedResultSet[] results, boolean isEarlier) throws IncompatibleTypesException, DateFormatException
		{
		sChannel.debug(sModule, "" + (isEarlier ? "earlier" : "later") + " result set length: " + results.length);

		HashMap<String, IGatherer> gathererMap = (isEarlier ? fEarlyGatherers : fLateGatherers);
		
		if(results == null)
			{
			sChannel.debug(sModule, "tried to aggregate a null result set");
			}
		
		for(int i = 0; i < results.length; i++)
			{
			if (isEarlier)
				System.out.print("+");
			else
				System.out.print("-");
			
			Iterator<MetricData> iterator = results[i].iterator();
			
			while(iterator.hasNext())
				{
				MetricData data = iterator.next();
				
				IGatherer gatherer = gathererMap.get(data.getMetric());
				
				if(gatherer == null)
					{
					int cookie = data.getMetric().getAttributeType();
					
					long when = -1; 
					if(isEarlier)
						{
						when = sEarlierPicker.getDateTimeSelector().getDate().getTime();
						}
					else
						{
						when = sLaterPicker.getDateTimeSelector().getDate().getTime();
						}
					
					gatherer = GathererFactory.newTimeAggregatingGathererByType(cookie, when);
					
					String pathAndAttribute = data.getMetric().getMetricPrefix().getAgentMetricPrefix().getPathURL() + ":" + data.getMetric().getAgentMetric().getAttributeName();
					//sChannel.debug(sModule, "caching " + pathAndAttribute);
					//sChannel.debug(sModule, "data.getAgentMetric().getAttributeURL()" + data.getAgentMetric().getAttributeURL());
					
					gathererMap.put(pathAndAttribute, gatherer);					
					}
				gatherer.acceptValueNonSync(data.getDataValue());
				}
			
			}
		}

	public void printResultSet(ITimeslicedResultSet[] timelicedResultSetArray)
		{

		for(int i = 0; i < timelicedResultSetArray.length; i++)
			{
			Iterator<MetricData> itr = timelicedResultSetArray[i].iterator();
			
			while(itr.hasNext())
				{
				MetricData metricData = itr.next();
				
				IMetricDataValue dataValue = metricData.getDataValue();
				String result = metricData.getMetric().getName() + " @ " + dataValue.getStopTimestampInMillis() + " = " + dataValue.getValueAsString();
				sChannel.debug(sModule, result);
				}
			}
		
		}

	// ruthlessly stolen from com/wily/introscope/console/thornhill/ui/typeviewer/SearchTypeViewer.java
	public IMetricSpecifier getQuerySpecifier(IThornhillItem item, AgentName agentName) throws InvalidRegexException
		{
		IMetricSpecifier spec = null;

		String regex = getQueryString(item);
		
		regex = RegularExpressionBuilder.convertClearTextToRegExp(regex);
		
		if(regex.indexOf(":") == -1)
			{
			regex += ".*"; // wildcard everything that's not a metric
			}
		
		sChannel.debug(sModule, "regex: " + regex);	
		IDomainSpecifier domainSpec = DomainSpecifier.getMatchAnyDomainSpecifier();
		sChannel.debug(sModule, "domainSpec: " + domainSpec);
		IAgentNameSpecifierIgnoringDomain agentSpec = AgentNameSpecifierIgnoringDomain.getAgentNameSpecifier(agentName);
		sChannel.debug(sModule, "agentSpec: " + agentSpec);
		IAgentMetricSpecifier partialSpec = AgentMetricSpecifier.getAgentMetricSpecifier(regex);
		sChannel.debug(sModule, "partialSpec: " + partialSpec.toString());
		
		// make the search case insensitive
		partialSpec.setCaseSensitive(false);
	
		// only include numerical metrics in this query (TT #23236)
		partialSpec = new BooleanAndAgentMetricSpecifier (partialSpec, new AnyNumericalAgentMetricSpecifier());
	
		spec = new MetricSpecifier (domainSpec, agentSpec, partialSpec);
	
		sChannel.debug(sModule, "query: " + spec.toString());
		return spec;
		}

	public String getQueryString(IThornhillItem item)
		{
		String result = "";
		
		if(item instanceof ThornhillAgent)
			{
			//result; // no change necessary; wild card added by caller
			}
		else if(item instanceof ThornhillMetricPath)
			{
			ThornhillMetricPath path = (ThornhillMetricPath) item;
			result = path.getAgentMetricPrefix().getPathURL();
			}
		else if(item instanceof ThornhillMetric)
			{
			ThornhillMetric metric = (ThornhillMetric) item;
			result = metric.getAgentMetric().getAgentMetricPrefix() + ":" + metric.getSimpleName();
			}
		
		sChannel.debug(item.getFullName() + " query string: " + result);
		return result;
		}

	public long 
	parseDuration(String selectedItem)
		{
		
		String[] split = selectedItem.split(" ");
		sChannel.debug(sModule, "selectedItem: " + selectedItem + " split[0]: " + split[0] + " split[1]: " + split[1]);
		
		long number = Long.parseLong(split[0]);
		
		long unitInMillis = -1;
		if (split[1].equalsIgnoreCase("Minute") || split[1].equalsIgnoreCase("Minutes"))
			{
			unitInMillis = 60 * 1000;
			}
		else if(split[1].equalsIgnoreCase("Hour") || split[1].equalsIgnoreCase("Hours"))
			{
			unitInMillis = 60 * 1000 * 60;
			}
		else if(split[1].equalsIgnoreCase("Day") || split[1].equalsIgnoreCase("Days"))
			{
			unitInMillis = 60 * 1000 * 60 * 24;
			}
		else
			{
			sChannel.debug(sModule, "Units not recognized" + split[1]);
			}
		
		return number * unitInMillis;
		}
	}
